@@ -2,7 +2,7 @@
 #
 # Ein Beispiel der DANTE-Edition
 #
-## Beispiel 12-04-1 auf Seite 600.
+## Beispiel 12-04-3 auf Seite 602.
 #
 # Copyright (C) 2016 Herbert Voss
 #
@@ -12,11 +12,11 @@
 #
 # See http://www.latex-project.org/lppl.txt for details.
 #
-# Running otfinfo on Linux
+# Running fc-list on Linux
 #START
 
 #STOP
 #
 #CODE->
-otfinfo -i `kpsewhich ComicNeue_Regular.otf`
+fc-list Cambria
 #<-CODE
