@@ -70,7 +70,7 @@ sub makevf_body {
     my ($face, $dir, $lang)=@_;
 
     my @exist_head=@{$MakeSPList::r_exist_head->{$lang}};
-    foreach $first_hex (0x1b, 0x1d .. 0x2b, 0x2f) { # U+1Cxxx, U+2[CDE]xxx : not defined yet
+    foreach $first_hex (0x1b, 0x1d .. 0x2f) { # U+1Cxxx : not defined yet
 	next if (!$exist_head[$first_hex]);
 
 	$HEX = sprintf("%X", $first_hex);
@@ -95,7 +95,7 @@ sub make_uni_vf_body {
     my $lang = join ',', @ln;
 
     my @exist_head=@{$MakeSPList::r_exist_head->{$lang}};
-    foreach $first_hex (0x1b, 0x1d .. 0x2b, 0x2f) { # U+1Cxxx, U+2[CDE]xxx : not defined yet
+    foreach $first_hex (0x1b, 0x1d .. 0x2f) { # U+1Cxxx : not defined yet
 	next if (!$exist_head[$first_hex]);
 
 	$HEX = sprintf("%X", $first_hex);
