@@ -152,6 +152,7 @@ sub glue_kern{
 	print  JPL "   )\n";
 }
 sub chars_in_type_jis{
+	my ($type3add) = $ucs ? "U00B7" : "";
 	my ($type5) = $ucs ? "— ― … ‥" : "— … ‥";
 print JPL <<END_OF_DATA;
 (CHARSINTYPE O 1
@@ -162,6 +163,7 @@ print JPL <<END_OF_DATA;
    )
 (CHARSINTYPE O 3
    ・ ： ； 
+   $type3add 
    )
 (CHARSINTYPE O 4
    。 ． 
