@@ -2,7 +2,7 @@
 This file is part of the `arabluatex' package
 
 ArabLuaTeX -- Processing ArabTeX notation under LuaLaTeX
-Copyright (C) 2016--2017  Robert Alessi
+Copyright (C) 2016--2018  Robert Alessi
 
 Please send error reports and suggestions for improvements to Robert
 Alessi <alessi@robertalessi.net>
@@ -46,7 +46,33 @@ local function unprotectarb(str)
 return str
 end
 
+brkcmds = {}
+
+function mkarbbreak(str)
+   str = str ..","
+   str = string.gsub(str, "%s+", "")
+   local fieldstart = 1
+   repeat
+      local nexti = string.find(str, "%,", fieldstart)
+      table.insert(brkcmds, string.sub(str, fieldstart, nexti-1))
+      fieldstart = nexti +1
+   until fieldstart > string.len(str)
+   return brkcmds
+end
+
 local function breakcmd(str)
+   -- user commands
+   if next(brkcmds) == nil then
+      -- nothing to do
+   else
+      for i = 1,#brkcmds do
+	 str = string.gsub(str, "\\"..brkcmds[i].."%s?(%b{})",
+		function(body)
+		   body = string.sub(body, 2, -2)
+		   return string.format("}\\"..brkcmds[i].."{%s}\\arb{", body)
+	 end)
+      end
+   end
    -- process \item[], then \item[]
 	str = string.gsub(str, "\\(item.?)(%b[])",
 	  function(tag, body)
@@ -73,6 +99,13 @@ local function breakcmd(str)
 		body = string.sub(body, 2, -2)
 		return string.format("}\\%s{%s}\\arb{", tag, body)
 		end)
+   -- \textcolor
+   str = string.gsub(str, "\\(textcolor.?)(%b{})(%b{})",
+    function(tag, bodycolor, bodytext)
+    bodycolor = string.sub(bodycolor, 2, -2)
+    bodytext = string.sub(bodytext, 2, -2)
+    return string.format("}\\%s{%s}{\\arb{%s}}\\arb{", tag, bodycolor, bodytext)
+   end)
    -- Footnote
    str = string.gsub(str, "\\(Footnote.-)(%b{})",
 		function(tag, body)
@@ -157,7 +190,12 @@ local function takeoutabjad(str)
 end
 
 local function takeoutcapetc(str)
-   str = string.gsub(str, "(\\cap.?)(%b{})", function(tag, body)
+   str = string.gsub(str, "(\\arb.?%[trans%])(%b{})", function(tag, body)
+			body = string.sub(body, 2, -2)
+			body = string.gsub(body, "(\\uc%s?)(%b{})", "\\Uc%2")
+			return string.format("%s{%s}", tag, body)
+   end)
+   str = string.gsub(str, "(\\uc.?)(%b{})", function(tag, body)
 			body = string.sub(body, 2, -2)
 			return string.format("%s", body)
    end)
@@ -323,8 +361,14 @@ local function fullvoceasy(str, rules)
 	 inside = string.gsub(inside, digraphsfv[i].a, digraphsfv[i].b)
       end
    end
-   for i = 1,#singlefveasy do
-      inside = string.gsub(inside, singlefveasy[i].a, singlefveasy[i].b)
+   if rules == "nosukun" then
+      for i = 1,#singlefveasy do
+	 inside = string.gsub(inside, singlefveasy[i].a, singlefveasy[i].b)
+      end
+   else
+      for i = 1,#singlefv do
+	 inside = string.gsub(inside, singlefv[i].a, singlefv[i].b)
+      end
    end
    for i = 1,#longv do
       inside = string.gsub(inside, longv[i].a, longv[i].b)
@@ -494,6 +538,41 @@ local function transloc(str)
 return str
 end
 
+local function transarabica(str)
+   str = string.gsub(str, "\\arb(%b{})", function(inside)
+   inside = string.sub(inside, 2, -2)
+   for i = 1,#hamzatrarabica do
+      inside = string.gsub(inside, hamzatrarabica[i].a, hamzatrarabica[i].b)
+   end
+   for i = 1,#tanwintrloc do
+      inside  = string.gsub(inside, tanwintrloc[i].a, tanwintrloc[i].b)
+   end
+   for i = 1,#trigraphstrarabica do
+      inside = string.gsub(inside, trigraphstrarabica[i].a, trigraphstrarabica[i].b)
+   end
+   for i = 1,#digraphstrarabica do
+      inside = string.gsub(inside, digraphstrarabica[i].a, digraphstrarabica[i].b)
+   end
+   for i = 1,#singletrarabica do
+      inside = string.gsub(inside, singletrarabica[i].a, singletrarabica[i].b)
+   end
+   for i = 1,#longvtrarabica do
+      inside = string.gsub(inside, longvtrarabica[i].a, longvtrarabica[i].b)
+   end
+   for i = 1,#shortvtrloc do
+      inside = string.gsub(inside, shortvtrloc[i].a, shortvtrloc[i].b)
+   end
+   for i = 1,#punctuationtr do
+      inside = string.gsub(inside, punctuationtr[i].a, punctuationtr[i].b)
+   end
+   for i = 1,#nulltr do
+      inside = string.gsub(inside, nulltr[i].a, nulltr[i].b)
+   end
+   return string.format("\\txtrans{%s}", inside)
+   end)
+return str
+end
+
 local function processbuckw(str)
    str = string.gsub(str, "\\arb(%b{})", function(inside)
    inside = string.sub(inside, 2, -2)
@@ -579,30 +658,65 @@ function processtrans(str, mode, rules, scheme)
       str = transdmg(str, rules)
    elseif mode == "loc" then
       str = transloc(str)
+   elseif mode == "arabica" then
+      str = transarabica(str)
    end
    str = unprotectarb(str)
 return str
 end
 
-function cap(str)
+function uc(str)
    str = string.gsub(str, "(\\txtrans.?)(%b{})", function(tag, body)
 			body = string.sub(body, 2, -2)
 			return string.format("%s", body)
    end)
-   if string.find(str, "[^\\]%-['`ʾʿ]") then
-      str = string.gsub(str, "([^\\]%-['`])", "%1\\MakeUppercase ")
-      str = string.gsub(str, "([^\\]%-ʿ)", "%1\\MakeUppercase ")
-      str = string.gsub(str, "([^\\]%-ʾ)", "%1\\MakeUppercase ")
-   elseif string.find(str, "[^\\]%-[^'`ʾʿ]") then
-      str = string.gsub(str, "([^\\]%-)", "%1\\MakeUppercase ")
-   elseif string.find(str, "^['`ʾʿ]") then
-      str = string.gsub(str, "^(['`])", "%1\\MakeUppercase ")
-      str = string.gsub(str, "^(ʿ)", "%1\\MakeUppercase ")
-      str = string.gsub(str, "^(ʾ)", "%1\\MakeUppercase ")
-   else
-      str = "\\MakeUppercase "..str
+   -- Allah and ibn
+   str = string.gsub(str, "(al%-lāh)([uai]?)", "{Allāh%2}")
+   str = string.gsub(str, "([%'%-]?)(l%-lāh)([uai]?)", "%1{Llāh%3}")
+   str = string.gsub(str, "(%s[%(%<%[]?)([i%']?b[n%.])", "%1{%2}")
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^([%S]-%-[`']?)"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
    end
-return "\\txtrans{"..str.."}"
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)([%S]-%-[`']?)"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^([%S]-%-ʿ)"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)([%S]-%-ʿ)"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^([%S]-%-ʾ)"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)([%S]-%-ʾ)"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^([`'])"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)([`'])"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^(ʾ)"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)(ʾ)"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^(ʿ)"..lcuc[i].a, "{%1"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)(ʿ)"..lcuc[i].a, "%1{%2"..lcuc[i].b.."}")
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "^"..lcuc[i].a, lcuc[i].b)
+   end
+   for i = 1,#lcuc do
+      str = string.gsub(str, "(%s[%(%<%[]?)"..lcuc[i].a, "%1"..lcuc[i].b)
+   end
+   return "\\txtrans{"..str.."}"
 end
 
 -- this function is adapted from an 'obsolete project' of Khaled
@@ -638,3 +752,12 @@ function abraces(str)
    end
    return str
 end
+
+function aemph(str)
+   if tex.textdir == "TRT" then
+      str = "$\\overline{\\text{"..str.."}}$"
+   elseif tex.textdir == "TLT" then
+      str = "$\\underline{\\text{"..str.."}}$"
+   end
+   return str
+end
