@@ -2,7 +2,7 @@
 This file is part of the `arabluatex' package
 
 ArabLuaTeX -- Processing ArabTeX notation under LuaLaTeX
-Copyright (C) 2016--2017  Robert Alessi
+Copyright (C) 2016--2018  Robert Alessi
 
 Please send error reports and suggestions for improvements to Robert
 Alessi <alessi@robertalessi.net>
@@ -170,6 +170,12 @@ buckwalter = {
 hamza = {
    -- next line for ʾiʿrāb hyphen
    {a="(')(%-)([uaiUAI])", b="%1%3"},
+   -- next lines for ʾalif alone
+   {a="(%.A)([uai]?)l%-(%^n)", b="ا%2ل%3"}, --additional (^n is lunar)
+   {a="([%(%[%|%<%s%-o])(%.A)([uai]?)l%-(%^n)", b="%1ا%3%4"}, --additional (^n is lunar) --p
+   {a="(%.A)([uai]?)l%-([%_%^%.]?[tdrzsnc])", b="ا%2ل%3%3"},
+   {a="([%(%[%|%<%s%-o])(%.A)([uai]?)l%-([%_%^%.]?[tdrzsnc])", b="%1ا%3ل%4%4"}, --p
+   {a="%.A", b="ا"},
    -- hard coded hamza
    {a="|\"'", b="ء"},
    {a="A\"'", b="آ"},
@@ -188,10 +194,15 @@ hamza = {
    {a="%'%_I", b="إI"},
    -- taḫfīfu 'l-hamza
    {a="'u'([^uaiUAI])", b="أU%1"},
-   {a="'i'([^uaiUAI])", b="إI%1"},   
+   {a="'i'([^uaiUAI])", b="إI%1"},
+   {a="([wf]a)%-\'([^uaiUAIl][^%-])", b="%1أْ%2"},
+   {a="^u'([^uaiUAI])", b="اU%1"},
+   {a="([^uaiUAIYN][%s%(%[%<])u'([^uaiUAI])", b="%1اU%2"},
+   {a="^i'([^uaiUAI])", b="اI%1"},
+   {a="([^uaiUAIYN][%s%(%[%<])i'([^uaiUAI])", b="%1اI%2"},
    -- madda (historic writing below)
    {a="'a'([^uaiUAI])", b="آ%1"},
-   {a="([^uiUI])\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
+   {a="([^uiyUI])\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
    {a="^\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="آ%1"},
    {a="(%W)\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
    {a="(A)(')(uN?%p?)$", b="aآء%3"},
@@ -214,20 +225,25 @@ hamza = {
    -- then the 'initial' rules for the remaining cases
    {a="^(')([ua])", b="أ%2"},
    {a="^(')(i)", b="إ%2"},
-   {a="(%W)(')([ua])", b="%1أ%3"},
-   {a="(%W)(')(i)", b="%1إ%3"},
+   -- consider replacing initial %W with [%s%(%[%<%-]:
+   --   {a="(%W)(')([ua])", b="%1أ%3"},
+   --   {a="(%W)(')(i)", b="%1إ%3"},
+   {a="([%s%(%[%<%-])(')([ua])", b="%1أ%3"},
+   {a="([%s%(%[%<%-])(')(i)", b="%1إ%3"},
    -- final
-   -- ^say'aN and .zim'aN are special orthographies
-   {a="(%^say)(%')(aN)", b="%1ئ%3"},
-   {a="(.zi?m)(%')(aN)", b="%1ئ%3"},
-   {a="(')(I%p?)$", b="إ%2"}, -- affix pr. 1st person sg.
-   {a="(')(I)(%p?%s)", b="إ%2%3"}, -- affix pr. 1st person sg.
-   {a="([^uai])(')([uai]N?%p?)$", b="%1ء%3"},
-   {a="([^uai])(')([uai]N?)(%p?%s)", b="%1ء%3%4"},
+   -- mi'aT is special orthography (unlike ^say'aN and .zim'aN):
+   -- {a="(%^sa%.?[yY])(\"?%|?)(%')(aN)", b="%1%2ئ%4"}, --new
+   -- {a="(.zi?m)(%')(aN)", b="%1ئ%3"}, --new
+   {a="(mi)(%')(a[Tt])", b="%1ائ%3"},
+   {a="(mi)(%')(aN%_?[AY])", b="%1أ%3"},
+   -- final hamzah is on the line after a letter of prolongation or a
+   -- consonant with sukūn
+   {a="([^Auai])(')(\"?[uai]N?)(%p?)$", b="%1ء%3%4"}, --new
+   {a="([^Auai])(')(\"?[uai]N?)(%p?%s)", b="%1ء%3%4"},
 -- u
    {a="(u)(')([uai]N?%p?)$", b="%1ؤ%3"},
    {a="(u)(')([uai]N?)(%p?%s)", b="%1ؤ%3%4"},
-   {a="(u)('%p?)$", b="%1ؤ"},
+   {a="(u)(')(%p?)$", b="%1ؤ%3"},
    {a="(u)(')(%p?%s)", b="%1ؤ%3"},
 -- a
    {a="(a)(')(A%p?)$", b="%1آ"},
@@ -240,27 +256,27 @@ hamza = {
    {a="(a)(')(aN)(%p?%s)", b="%1أً%4"},
    {a="(a)(')([i]N?%p?)$", b="%1إ%3"},
    {a="(a)(')([i]N?)(%p?%s)", b="%1إ%3%4"},
-   {a="(a)('%p?)$", b="%1أ"},
+   {a="(a)(')(%p?)$", b="%1أ%3"},
    {a="(a)(')(%p?%s)", b="%1أ%3"},
 -- i
    {a="(i)(')([uai]N?%p?)$", b="%1ئ%3"},
    {a="(i)(')([uai]N?)(%p?%s)", b="%1ئ%3%4"},
-   {a="(i)('%p?)$", b="%1ئ"},
+   {a="(i)(')(%p?)$", b="%1ئ%3"},
    {a="(i)(')(%p?%s)", b="%1ئ%3"},
 --
    -- middle
-   {a="(U)(')", b="%1ء"},
-   {a="([Iy])(')", b="%1ئ"},
+   {a="([UIwy])(')", b="%1ء"}, --new
+   -- {a="([Iy])(')", b="%1ئ"}, -- included in the above line
    {a="([^uai])(')([uU])", b="%1ؤ%3"},
-   {a="([^uai])(')([aA])", b="%1أ%3"},
+   {a="([^uai])(')(%_?[aAY])", b="%1أ%3"},
    {a="([^uai])(')([iI])", b="%1ئ%3"},
    {a="(u)(')([uU])", b="%1ؤ%3"},
-   {a="(u)(')([aA])", b="%1ؤ%3"},
+   {a="(u)(')(%_?[aAY])", b="%1ؤ%3"},
    {a="(u)(')([iI])", b="%1ئ%3"},
-   {a="(a)(')([aA])", b="%1أ%3"},
+   {a="(a)(')(%_?[aAY])", b="%1أ%3"},
    {a="(a)(')([uU])", b="%1ؤ%3"},
    {a="(a)(')([iI])", b="%1ئ%3"},
-   {a="(i)(')([aA])", b="%1ئ%3"},
+   {a="(i)(')(%_?[aAY])", b="%1ئ%3"},
    {a="(i)(')([uU])", b="%1ئ%3"},
    {a="(i)(')([iI])", b="%1ئ%3"},
    {a="(a)(')([^uaiUAI])", b="%1أ%3"},
@@ -271,6 +287,12 @@ hamza = {
 hamzaeasy = { -- differences marked below with 'easy'
    -- next line for ʾiʿrāb hyphen
    {a="(')(%-)([uaiUAI])", b="%1%3"},
+   -- next lines for ʾalif alone (easy)
+   {a="(%.A)([uai]?)l%-(%^n)", b="ا%2ل%3"}, --additional (^n is lunar)
+   {a="([%(%[%|%<%s%-o])(%.A)([uai]?)l%-(%^n)", b="%1ا%3%4"}, --additional (^n is lunar) --p
+   {a="(%.A)([uai]?)l%-([%_%^%.]?[tdrzsnc])", b="ا%2ل%3"},
+   {a="([%(%[%|%<%s%-o])(%.A)([uai]?)l%-([%_%^%.]?[tdrzsnc])", b="%1ا%3ل%4"}, --p
+   {a="%.A", b="ا"},
    -- hard coded hamza
    {a="|\"'", b="ء"},
    {a="A\"'", b="آ"},
@@ -289,10 +311,15 @@ hamzaeasy = { -- differences marked below with 'easy'
    {a="%'%_I", b="إI"},
    -- taḫfīfu 'l-hamza
    {a="'u'([^uaiUAI])", b="أU%1"},
-   {a="'i'([^uaiUAI])", b="إI%1"},   
+   {a="'i'([^uaiUAI])", b="إI%1"},
+   {a="([wf]a)%-\'([^uaiUAIl][^%-])", b="%1أْ%2"},
+   {a="^u'([^uaiUAI])", b="اU%1"},
+   {a="([^uaiUAIYN][%s%(%[%<])u'([^uaiUAI])", b="%1اU%2"},
+   {a="^i'([^uaiUAI])", b="اI%1"},
+   {a="([^uaiUAIYN][%s%(%[%<])i'([^uaiUAI])", b="%1اI%2"},
    -- madda (historic writing below)
    {a="'a'([^uaiUAI])", b="آ%1"},
-   {a="([^uiUI])\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
+   {a="([^uiyUI])\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
    {a="^\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="آ%1"},
    {a="(%W)\'a?A([%_%^%.]?[%`%'btjghxdrzsfqklmnywAY])", b="%1آ%2"},
    --easy (begin)
@@ -317,20 +344,35 @@ hamzaeasy = { -- differences marked below with 'easy'
    -- then the 'initial' rules for the remaining cases
    {a="^(')([ua])", b="أ%2"},
    {a="^(')(i)", b="إ%2"},
-   {a="(%W)(')([ua])", b="%1أ%3"},
-   {a="(%W)(')(i)", b="%1إ%3"},
+   -- consider replacing initial %W with [%s%(%[%<%-]:
+   --   {a="(%W)(')([ua])", b="%1أ%3"},
+   --   {a="(%W)(')(i)", b="%1إ%3"},
+   {a="([%s%(%[%<%-])(')([ua])", b="%1أ%3"},
+   {a="([%s%(%[%<%-])(')(i)", b="%1إ%3"},
    -- final
-   -- ^say'aN and .zim'aN are special orthographies
-   {a="(%^say)(%')(aN)", b="%1ئ%3"},
-   {a="(.zi?m)(%')(aN)", b="%1ئ%3"},
-   {a="(')(I%p?)$", b="إ%2"}, -- affix pr. 1st person sg.
-   {a="(')(I)(%p?%s)", b="إ%2%3"}, -- affix pr. 1st person sg.
-   {a="([^uai])(')([uai]N?%p?)$", b="%1ء%3"},
-   {a="([^uai])(')([uai]N?)(%p?%s)", b="%1ء%3%4"},
+   -- mi'aT is special orthography (unlike ^say'aN and .zim'aN)
+   -- {a="(%^sa%.?[yY])(\"?%|?)(%')(aN)", b="%1%2ئ%4"}, --new
+   -- {a="(.zi?m)(%')(aN)", b="%1ئ%3"}, --new
+   {a="(mi)(%')(a[Tt])", b="%1ائ%3"},
+   {a="(mi)(%')(aN%_?[AY])", b="%1أ%3"},
+   -- easy (begin)
+   -- The Munjid says that such words as radI'aN do not have the
+   -- hamzah alone on the line, so take out the following two lines
+   -- (final hamzah is on the line after a letter of prolongation or a
+   -- consonant with sukūn)
+   -- {a="([^Auai])(')(\"?[uai]N?)(%p?)$", b="%1ء%3%4"}, --new
+   -- {a="([^Auai])(')(\"?[uai]N?)(%p?%s)", b="%1ء%3%4"},
+   {a="([^Auai])(')(\"?aN)(%p?)$", b="%1ئ%3%4"}, --new
+   {a="([^Auai])(')(\"?aN)(%p?%s)", b="%1ئ%3%4"}, --new
+   {a="([^uai])(')(\"?a)(%p?)$", b="%1ء%3%4"}, --new
+   {a="([^uai])(')(\"?a)(%p?%s)", b="%1ء%3%4"}, --new
+   {a="([^uai])(')(\"?[ui]N?)(%p?)$", b="%1ء%3%4"}, --new
+   {a="([^uai])(')(\"?[ui]N?)(%p?%s)", b="%1ء%3%4"}, --new
+   --easy (end)
 -- u
    {a="(u)(')([uai]N?%p?)$", b="%1ؤ%3"},
    {a="(u)(')([uai]N?)(%p?%s)", b="%1ؤ%3%4"},
-   {a="(u)('%p?)$", b="%1ؤ"},
+   {a="(u)(')(%p?)$", b="%1ؤ%3"},
    {a="(u)(')(%p?%s)", b="%1ؤ%3"},
 -- a
    {a="(a)(')(A%p?)$", b="%1آ"},
@@ -343,27 +385,27 @@ hamzaeasy = { -- differences marked below with 'easy'
    {a="(a)(')(aN)(%p?%s)", b="%1أً%4"},
    {a="(a)(')([i]N?%p?)$", b="%1إ%3"},
    {a="(a)(')([i]N?)(%p?%s)", b="%1إ%3%4"},
-   {a="(a)('%p?)$", b="%1أ"},
+   {a="(a)(')(%p?)$", b="%1أ%3"},
    {a="(a)(')(%p?%s)", b="%1أ%3"},
 -- i
    {a="(i)(')([uai]N?%p?)$", b="%1ئ%3"},
    {a="(i)(')([uai]N?)(%p?%s)", b="%1ئ%3%4"},
-   {a="(i)('%p?)$", b="%1ئ"},
+   {a="(i)(')(%p?)$", b="%1ئ%3"},
    {a="(i)(')(%p?%s)", b="%1ئ%3"},
 --
    -- middle
-   {a="(U)(')", b="%1ء"},
-   {a="([Iy])(')", b="%1ئ"},
+   {a="([Uw])(')", b="%1ء"}, --new
+   {a="([Iy])(')", b="%1ئ"}, --easy
    {a="([^uai])(')([uU])", b="%1ؤ%3"},
-   {a="([^uai])(')([aA])", b="%1أ%3"},
+   {a="([^uai])(')(%_?[aAY])", b="%1أ%3"},
    {a="([^uai])(')([iI])", b="%1ئ%3"},
    {a="(u)(')([uU])", b="%1ؤ%3"},
-   {a="(u)(')([aA])", b="%1ؤ%3"},
+   {a="(u)(')(%_?[aAY])", b="%1ؤ%3"},
    {a="(u)(')([iI])", b="%1ئ%3"},
-   {a="(a)(')([aA])", b="%1أ%3"},
+   {a="(a)(')(%_?[aAY])", b="%1أ%3"},
    {a="(a)(')([uU])", b="%1ؤ%3"},
    {a="(a)(')([iI])", b="%1ئ%3"},
-   {a="(i)(')([aA])", b="%1ئ%3"},
+   {a="(i)(')(%_?[aAY])", b="%1ئ%3"},
    {a="(i)(')([uU])", b="%1ئ%3"},
    {a="(i)(')([iI])", b="%1ئ%3"},
    {a="(a)(')([^uaiUAI])", b="%1أ%3"},
@@ -387,7 +429,7 @@ tanwin = {
    {a="%-?(aN)(_A)(%s)([rlmnwy])", b="ًى%3%4%4"},
    {a="%-?(aN)(Y)(%s)([rlmnwy])", b="ًى%3%4%4"},
    {a="(T)%-?(aN)(%s)([rlmnwy])", b="%1ً%3%4%4"},
-   {a="(ء)%-?(aN)(%s)([rlmnwy])", b="%1ً%3%4%4"},
+   {a="(ء)%-?(aN)(%s)([rlmnwy])", b="%1%2%3%4%4"}, --new
    {a="([^TA])%-?(aN)(%s)([rlmnwy])", b="%1ًا%3%4%4"},
    {a="%-?(iNI?)(%s)([rlmnwy])", b="ٍ%2%3%3"},
    -- assimilations (end)
@@ -399,7 +441,7 @@ tanwin = {
    {a="%-?(\"aN)(_A)", b="ى"},
    {a="%-?(\"aN)(Y)", b="ى"},
    {a="(T)%-?(\"aN)", b="%1"},
-   {a="(ء)%-?(\"aN)", b="%1"},
+   {a="([اآ])(ء)%-?(\"aN)", b="%1%2"}, --new
    {a="([^TA])%-?(\"aN)", b="%1ا"},
    {a="%-?(\"iNI?)", b=""},
    -- quoted tanwīn (end)
@@ -408,7 +450,7 @@ tanwin = {
    {a="%-?(aN)(_A)", b="ًى"},
    {a="%-?(aN)(Y)", b="ًى"},
    {a="(T)%-?(aN)", b="%1ً"},
-   {a="(ء)%-?(aN)", b="%1ً"},
+   {a="([اآ])(ء)%-?(aN)", b="%1%2ً"}, --new
    {a="([^TA])%-?(aN)", b="%1ًا"},
    {a="%-?(iNI?)", b="ٍ"}
 }
@@ -429,7 +471,7 @@ tanwineasy = { -- 'easy' requires some lines to be taken out:
 --   {a="%-?(aN)(_A)(%s)([rlmnwy])", b="ًى%3%4%4"},
 --   {a="%-?(aN)(Y)(%s)([rlmnwy])", b="ًى%3%4%4"},
 --   {a="(T)%-?(aN)(%s)([rlmnwy])", b="%1ً%3%4%4"},
---   {a="(ء)%-?(aN)(%s)([rlmnwy])", b="%1ً%3%4%4"},
+--   {a="(ء)%-?(aN)(%s)([rlmnwy])", b="%1%2%3%4%4"}, --new
 --   {a="([^TA])%-?(aN)(%s)([rlmnwy])", b="%1ًا%3%4%4"},
 --   {a="%-?(iNI?)(%s)([rlmnwy])", b="ٍ%2%3%3"},
    -- assimilations (end)
@@ -441,7 +483,7 @@ tanwineasy = { -- 'easy' requires some lines to be taken out:
    {a="%-?(\"aN)(_A)", b="ى"},
    {a="%-?(\"aN)(Y)", b="ى"},
    {a="(T)%-?(\"aN)", b="%1"},
-   {a="(ء)%-?(\"aN)", b="%1"},
+   {a="([اآ])(ء)%-?(\"aN)", b="%1%2"}, --new
    {a="([^TA])%-?(\"aN)", b="%1ا"},
    {a="%-?(\"iNI?)", b=""},
    -- quoted tanwīn (end)
@@ -450,7 +492,7 @@ tanwineasy = { -- 'easy' requires some lines to be taken out:
    {a="%-?(aN)(_A)", b="ًى"},
    {a="%-?(aN)(Y)", b="ًى"},
    {a="(T)%-?(aN)", b="%1ً"},
-   {a="(ء)%-?(aN)", b="%1ً"},
+   {a="([اآ])(ء)%-?(aN)", b="%1%2ً"}, --new
    {a="([^TA])%-?(aN)", b="%1ًا"},
    {a="%-?(iNI?)", b="ٍ"}
 }
@@ -471,12 +513,16 @@ trigraphs = { -- trigraphs or more
    -- al- + lām
    {a="^(a)l%-(l)", b="ا%1ل%2%2"},
    {a="([%(%[%|%<%s%-o])(a)l%-(l)", b="%1ا%2ل%3%3"}, --p
-   -- al- + solar consonant
-   {a="^(a)l%-([%_%^%.]?[tdrzsn])", b="ا%1ل%2%2"},
-   {a="([%(%[%|%<%s%-o])(a)l%-([%_%^%.]?[tdrzsn])", b="%1ا%2ل%3%3"}, --p
-   -- assim. art. + solar consonant
-   {a="^(a)([%_%^%.]?[tdrzsn])%-", b="ا%1ل%2"},
-   {a="([%(%[%|%<%s%-o])(a)([%_%^%.]?[tdrzsn])%-", b="%1ا%2ل%3"}, --p
+   -- al- + solar consonant ('c' and '^n' are additional characters)
+   {a="^(a)l%-(%^n)", b="ا%1ل%2"}, -- ^n is lunar
+   {a="([%(%[%|%<%s%-o])(a)l%-(%^n)", b="%1ا%2ل%3"},-- ^n is lunar --p
+   {a="^(a)l%-([%_%^%.]?[tdrzsnc])", b="ا%1ل%2%2"},
+   {a="([%(%[%|%<%s%-o])(a)l%-([%_%^%.]?[tdrzsnc])", b="%1ا%2ل%3%3"}, --p
+   -- assim. art. + solar consonant ('c' and '^n' are additional characters)
+   {a="^(a)(%^n)%-", b="ا%1ل"}, -- ^n is lunar
+   {a="([%(%[%|%<%s%-o])(a)(%^n)%-", b="%1ا%2ل"},-- ^n is lunar --p
+   {a="^(a)([%_%^%.]?[tdrzsnc])%-", b="ا%1ل%2"},
+   {a="([%(%[%|%<%s%-o])(a)([%_%^%.]?[tdrzsnc])%-", b="%1ا%2ل%3"}, --p
    -- al- + initial unstable hamza
    {a="^(a)l%-(\")([uai])", b="ا%1ل%3ٱ"},
    {a="([%(%[%|%<%s%-o])(a)l%-(\")([uai])", b="%1ا%2ل%4ٱ"}, --p
@@ -498,15 +544,23 @@ trigraphs = { -- trigraphs or more
    -- art. with waṣla + lām
    {a="'l%-(l)", b="ال%1%1"},
    -- art. with waṣla + solar consonant
-   {a="'l%-([%_%^%.]?[tdrzsn])", b="ال%1%1"},
+   -- ('c' and '^n' are additional characters)
+   {a="'l%-(%^n)", b="ال%1"}, -- ^n is lunar
+   {a="'l%-([%_%^%.]?[tdrzsnc])", b="ال%1%1"},
    -- li-/la- + art. + lām
    {a="l([ai])%-l%-(l)", b="ل%1%2%2"},
-   -- assim. art. with waṣla + solar consonant
-   {a="'([%_%^%.]?[tdrzsn])%-", b="ال%1"},
+   -- assim. art. with waṣla + solar consonant ('c' and '^n' are
+   -- additional characters)   
+   {a="'(%^n)%-", b="ال"}, -- ^n is lunar
+   {a="'([%_%^%.]?[tdrzsnc])%-", b="ال%1"},
    -- li-/la- + art. + solar consonant is a special orthography
-   {a="l([ai])%-l%-([%_%^%.]?[tdrzsn])", b="ل%1ل%2%2"},
+   -- ('c' and '^n' are additional characters)
+   {a="l([ai])%-l%-(%^n)", b="ل%1ل%2"}, -- '^n' is lunar
+   {a="l([ai])%-l%-([%_%^%.]?[tdrzsnc])", b="ل%1ل%2%2"},
    -- li-/la + assim. art. + solar consonant is a special orthography
-   {a="l([ai])%-([%_%^%.]?[tdrzsn])%-([%_%^%.]?[tdrzsn])", b="ل%1ل%3%3"},
+   -- ('c' and '^n' are additional characters)
+   {a="l([ai])%-(%^n)%-(%^n)", b="ل%1ل%3"}, -- ^n is lunar
+   {a="l([ai])%-([%_%^%.]?[tdrzsnc])%-([%_%^%.]?[tdrzsnc])", b="ل%1ل%3%3"},
    -- art. with waṣla + initial unstable hamza
    {a="'l%-(\")([uai])", b="ال%2ٱ"},
    {a="'l%-([uai])", b="ال%1ا"},   
@@ -549,12 +603,17 @@ trigraphseasy = { -- differences marked below with 'easy'
    -- al- + lām (easy)
    {a="^(a)l%-(l)", b="ا%1ل%2"},
    {a="([%(%[%|%<%s%-o])(a)l%-(l)", b="%1ا%2ل%3"}, --p
-   -- al- + solar consonant (easy)
-   {a="^(a)l%-([%_%^%.]?[tdrzsn])", b="ا%1ل%2"},
-   {a="([%(%[%|%<%s%-o])(a)l%-([%_%^%.]?[tdrzsn])", b="%1ا%2ل%3"}, --p
-   -- assim. art. + solar consonant (easy)
-   {a="^(a)([%_%^%.]?[tdrzsn])%-", b="ا%1ل"},
-   {a="([%(%[%|%<%s%-o])(a)([%_%^%.]?[tdrzsn])%-", b="%1ا%2ل"}, --p
+   -- al- + solar consonant (easy) ('c' and '^n' are additional characters)
+   {a="^(a)l%-(%^n)", b="ا%1ل%2"}, -- ^n is lunar
+   {a="([%(%[%|%<%s%-o])(a)l%-(%^n)", b="%1ا%2ل%3"}, -- ^n is lunar --p
+   {a="^(a)l%-([%_%^%.]?[tdrzsnc])", b="ا%1ل%2"},
+   {a="([%(%[%|%<%s%-o])(a)l%-([%_%^%.]?[tdrzsnc])", b="%1ا%2ل%3"}, --p
+   -- assim. art. + solar consonant (easy) ('c' and '^n' are
+   -- additional characters)
+   {a="^(a)(%^n)%-", b="ا%1ل"}, -- ^n is lunar
+   {a="([%(%[%|%<%s%-o])(a)(%^n)%-", b="%1ا%2ل"}, -- ^n is lunar --p
+   {a="^(a)([%_%^%.]?[tdrzsnc])%-", b="ا%1ل"},
+   {a="([%(%[%|%<%s%-o])(a)([%_%^%.]?[tdrzsnc])%-", b="%1ا%2ل"}, --p
    -- al- + initial unstable hamza
    {a="^(a)l%-(\")([uai])", b="ا%1ل%3ٱ"},
    {a="([%(%[%|%<%s%-o])(a)l%-(\")([uai])", b="%1ا%2ل%4ٱ"}, --p
@@ -576,15 +635,23 @@ trigraphseasy = { -- differences marked below with 'easy'
    -- art. with waṣla + lām (easy)
    {a="'l%-(l)", b="ال%1"},
    -- art. with waṣla + solar consonant (easy)
-   {a="'l%-([%_%^%.]?[tdrzsn])", b="ال%1"},
+   -- ('c' and '^n' are additional characters)
+   {a="'l%-(%^n)", b="ال%1"}, -- ^n is lunar
+   {a="'l%-([%_%^%.]?[tdrzsnc])", b="ال%1"},
    -- li-/la- + art. + lām (easy)
    {a="l([ai])%-l%-(l)", b="ل%1%2"},
    -- assim. art. with waṣla + solar consonant (easy)
-   {a="'([%_%^%.]?[tdrzsn])%-", b="ال"},
+   -- ('c' and '^n' are additional characters)
+   {a="'(%^n)%-", b="ال"}, -- ^n is lunar
+   {a="'([%_%^%.]?[tdrzsnc])%-", b="ال"},
    -- li-/la- + art. + solar consonant is a special orthography (easy)
-   {a="l([ai])%-l%-([%_%^%.]?[tdrzsn])", b="ل%1ل%2"},
+   -- ('c' and '^n' are additional characters)
+   {a="l([ai])%-l%-(%^n)", b="ل%1ل%2"}, -- ^n is lunar
+   {a="l([ai])%-l%-([%_%^%.]?[tdrzsnc])", b="ل%1ل%2"},
    -- li-/la + assim. art. + solar consonant is a special orthography (easy)
-   {a="l([ai])%-([%_%^%.]?[tdrzsn])%-([%_%^%.]?[tdrzsn])", b="ل%1ل%3"},
+   -- ('c' and '^n' are additional characters)
+   {a="l([ai])%-(%^n)%-(%^n)", b="ل%1ل%3"}, -- ^n is lunar
+   {a="l([ai])%-([%_%^%.]?[tdrzsnc])%-([%_%^%.]?[tdrzsnc])", b="ل%1ل%3"},
    -- art. with waṣla + initial unstable hamza
    {a="'l%-(\")([uai])", b="ال%2ٱ"},
    {a="'l%-([uai])", b="ال%1ا"},
@@ -627,7 +694,7 @@ digraphs = {
    {a="(ay)(o)(\"?[uai])([%S]-o)", b="%1i"},
    {a="(ay)(%s)([%(%[%|%<]?)(\"?[uai])", b="%1i%2%3ا"}, --p
    -- hyphen + initial alif without hamza:
-   {a="([uai]%-)(\"?[uai])([%^%_%.%`]?)(%a)", b="%1ا%3%4"},
+   {a="([uai]%-)(\"?[uai])([%^%_%.%`]?)([%aإأؤئ])", b="%1ا%3%4"},
    -- initial alif without hamza
    {a="^([%(%[%|%<]?)(\"?[uai])", b="%1ا%2"}, --p
    -- initial alif without hamza
@@ -639,7 +706,9 @@ digraphs = {
    {a="ئئ", b="ئّ"},
    {a="bb", b="بّ"},
    {a="BB", b="ـّ"},
-   {a="(%_)([thd])([thd])", b="%1%2|%3"},
+   {a="([%_%^%.])([tghdsz])([tghdsz])", b="%1%2|%3"},
+   -- same as above for additional characters:
+   {a="([%_%^%.])([cn])([cn])", b="%1%2|%3"},
    {a="tt", b="تّ"},
    {a="%_t%_t", b="ثّ"},
    {a="jj", b="جّ"},
@@ -668,6 +737,15 @@ digraphs = {
    {a="hh", b="هّ"},
    {a="ww", b="وّ"},
    {a="yy", b="يّ"},
+   {a="%.y%.y", b="ىّ"},
+   -- additional characters + šaddah (begin)
+   {a="pp", b="پّ"},
+   {a="vv", b="ڤّ"},
+   {a="gg", b="گّ"},
+   {a="%^c%^c", b="چّ"},
+   {a="%^z%^z", b="ژّ"},
+   {a="%^n%^n", b="ڭّ"},
+   -- additional characters + šaddah (end)
    {a="_t", b="ث"},
    {a="%^g", b="ج"},
    {a="%.h", b="ح"},
@@ -679,6 +757,12 @@ digraphs = {
    {a="%.t", b="ط"},
    {a="%.z", b="ظ"},
    {a="%.g", b="غ"},
+   {a="%.y", b="ى"},
+   -- additional characters (begin)
+   {a="%^c", b="چ"},
+   {a="%^z", b="ژ"},
+   {a="%^n", b="ڭ"},
+   -- additional characters (end)
    {a="(U)(A)", b="%1ا"},
    {a="WA", b="وا"},
    {a="(a)W\"", b="%1وْا"},
@@ -716,6 +800,11 @@ single = {
    {a="w", b="و"},
    {a="y", b="ي"},
    {a="T", b="ة"},
+   -- additional characters (begin)
+   {a="p", b="پ"},
+   {a="v", b="ڤ"},
+   {a="g", b="گ"},
+   -- additional characters (end)
    {a="\"$", b="ْ"},
    {a="\"(%W)", b="ْ%1"},
    {a="\"([^uaiUAI])", b="ْ%1"},
@@ -731,6 +820,8 @@ longv = {
    {a="A", b="َا"},
    {a="U", b="ُو"},
    {a="I", b="ِي"},
+   {a="aY", b="aى"},
+   {a="iY", b="iى"},
    {a="Y", b="aى"}
 }
 
@@ -771,5 +862,6 @@ punctuation = {
 null = {
    {a="%|", b=""},
    {a="([^0-9])(%-)", b="%1"},
-   {a="o[%S]-o", b=""}
+   {a="o[%S]-o", b=""},
+   {a="[%^%_](.)", b=">??<%1"}
 }
