@@ -1,6 +1,6 @@
-# $Id: TLDownload.pm 39300 2016-01-07 02:48:24Z preining $
+# $Id: TLDownload.pm 42254 2016-10-11 01:22:02Z preining $
 # TeXLive::TLDownload.pm - module for abstracting the download modes
-# Copyright 2009-2014 Norbert Preining
+# Copyright 2009-2016 Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
@@ -8,9 +8,8 @@ package TeXLive::TLDownload;
 
 use TeXLive::TLUtils;
 use TeXLive::TLConfig;
-use File::Temp qw/tempfile/;
 
-my $svnrev = '$Revision: 39300 $';
+my $svnrev = '$Revision: 42254 $';
 my $_modulerevision;
 if ($svnrev =~ m/: ([0-9]+) /) {
   $_modulerevision = $1;
@@ -120,7 +119,7 @@ sub get_file {
   my $realout = $out;
   my ($outfh, $outfn);
   if ($out eq "|") {
-    ($outfh, $outfn) = tempfile(UNLINK => 1);
+    ($outfh, $outfn) = tl_tmpfile();
     $realout = $outfn;
   }
   my $response = $self->{'ua'}->get($url, ':content_file' => $realout);
@@ -134,7 +133,7 @@ sub get_file {
       return $outfh;
     }
   } else {
-    tlwarn("TLDownload::get_file: response error: "
+    debug("TLDownload::get_file: response error: "
             . $response->status_line . " (for $url)\n");
     $self->incr_errorcount;
     return;
