@@ -1,30 +1,25 @@
-# $Id: TLPOBJ.pm 29883 2013-04-13 05:32:44Z preining $
+# $Id: TLPOBJ.pm 46745 2018-02-26 18:16:54Z karl $
 # TeXLive::TLPOBJ.pm - module for using tlpobj files
-# Copyright 2007, 2008, 2009, 2010, 2011 Norbert Preining
+# Copyright 2007-2018 Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
 package TeXLive::TLPOBJ;
 
+my $svnrev = '$Revision: 46745 $';
+my $_modulerevision = ($svnrev =~ m/: ([0-9]+) /) ? $1 : "unknown";
+sub module_revision { return $_modulerevision; }
+
 use TeXLive::TLConfig qw($DefaultCategory $CategoriesRegexp 
                          $MetaCategoriesRegexp $InfraLocation 
                          $RelocPrefix $RelocTree);
-use TeXLive::TLUtils;
+use TeXLive::TLCrypto;
 use TeXLive::TLTREE;
+use TeXLive::TLUtils;
 
 our $_tmp;
 my $_containerdir;
 
-my $svnrev = '$Revision: 29883 $';
-my $_modulerevision;
-if ($svnrev =~ m/: ([0-9]+) /) {
-  $_modulerevision = $1;
-} else {
-  $_modulerevision = "unknown";
-}
-sub module_revision {
-  return $_modulerevision;
-}
 
 sub new {
   my $class = shift;
@@ -118,14 +113,23 @@ sub from_fh {
         # docfiles can have tags, but the parse_line function is so
         # time intense that we try to call it only when necessary
         if (defined $rest) {
-          my @words = &TeXLive::TLUtils::parse_line('\s+', 0, $rest);
-          for (@words) {
-            my ($k, $v) = split('=', $_, 2);
-            if ($k eq 'details' || $k eq 'language') {
-              $self->{'docfiledata'}{$f}{$k} = $v;
-            } else {
-              die "Unknown docfile tag: $line";
-            }
+          # parse_line has problems with double quotes in double quotes
+          # my @words = &TeXLive::TLUtils::parse_line('\s+', 0, $rest);
+          # do manual parsing
+          # this is not optimal, but since we support only two tags there
+          # are not so many cases
+          if ($rest =~ m/^details="(.*)"\s*$/) {
+            $self->{'docfiledata'}{$f}{'details'} = $1;
+          } elsif ($rest =~ m/^language="(.*)"\s*$/) {
+            $self->{'docfiledata'}{$f}{'language'} = $1;
+          } elsif ($rest =~ m/^language="(.*)"\s+details="(.*)"\s*$/) {
+            $self->{'docfiledata'}{$f}{'details'} = $2;
+            $self->{'docfiledata'}{$f}{'language'} = $1;
+          } elsif ($rest =~ m/^details="(.*)"\s+language="(.*)"\s*$/) {
+            $self->{'docfiledata'}{$f}{'details'} = $1;
+            $self->{'docfiledata'}{$f}{'language'} = $2;
+          } else {
+            tlwarn("$0: Unparsable tagging in TLPDB line: $line\n");
           }
         }
       } elsif ($lastcmd eq "binfiles") {
@@ -154,11 +158,16 @@ sub from_fh {
       }
     } elsif ($cmd eq 'containersize' || $cmd eq 'srccontainersize'
         || $cmd eq 'doccontainersize') {
-      $arg =~ /^[0-9]+$/ or die "Illegal size value: $line!";
+      $arg =~ /^[0-9]+$/ or die "Invalid size value: $line!";
       $self->{$cmd} = $arg;
     } elsif ($cmd eq 'containermd5' || $cmd eq 'srccontainermd5'
         || $cmd eq 'doccontainermd5') {
-      $arg =~ /^[a-f0-9]+$/ or die "Illegal md5 value: $line!";
+      $arg =~ /^[a-f0-9]{32}$/ or die "Invalid md5 value: $line!";
+      $self->{$cmd} = $arg;
+    } elsif ($cmd eq 'containerchecksum' || $cmd eq 'srccontainerchecksum'
+        || $cmd eq 'doccontainerchecksum') {
+      $arg =~ /^[a-f0-9]{$TeXLive::TLConfig::ChecksumLength}$/
+        or die "Invalid checksum value: $line!";
       $self->{$cmd} = $arg;
     } elsif ($cmd eq 'name') {
       $arg =~ /^([-.\w]+)$/ or die("Invalid name: $line!");
@@ -193,7 +202,7 @@ sub from_fh {
         $self->{'binsize'}{$arch} = $size;
       }
     } elsif ($cmd eq 'relocated') {
-      ($arg eq '0' || $arg eq '1') or die "Illegal value: $line!";
+      ($arg eq '0' || $arg eq '1') or die "Invalid value: $line!";
       $self->{'relocated'} = $arg;
     } elsif ($cmd eq 'catalogue') {
       $self->{'catalogue'} = $arg;
@@ -270,7 +279,7 @@ sub writeout {
   defined($self->{'shortdesc'}) && print $fd "shortdesc $self->{'shortdesc'}\n";
   defined($self->{'license'}) && print $fd "license $self->{'license'}\n";
   defined($self->{'relocated'}) && $self->{'relocated'} && print $fd "relocated 1\n";
-  # ugly hack to get rid of use FileHandle; see man perlform
+  # don't want to use FileHandle.pm; see man perlform
   #format_name $fd "multilineformat";
   select((select($fd),$~ = "multilineformat")[0]);
   $fd->format_lines_per_page (99999); # no pages in this format
@@ -299,21 +308,31 @@ sub writeout {
   if (defined($self->{'containermd5'})) {
     print $fd "containermd5 $self->{'containermd5'}\n";
   }
+  if (defined($self->{'containerchecksum'})) {
+    print $fd "containerchecksum $self->{'containerchecksum'}\n";
+  }
   if (defined($self->{'doccontainersize'})) {
     print $fd "doccontainersize $self->{'doccontainersize'}\n";
   }
   if (defined($self->{'doccontainermd5'})) {
     print $fd "doccontainermd5 $self->{'doccontainermd5'}\n";
   }
+  if (defined($self->{'doccontainerchecksum'})) {
+    print $fd "doccontainerchecksum $self->{'doccontainerchecksum'}\n";
+  }
   if (defined($self->{'docfiles'}) && (@{$self->{'docfiles'}})) {
     print $fd "docfiles size=$self->{'docsize'}\n";
     foreach my $f (sort @{$self->{'docfiles'}}) {
       print $fd " $f";
       if (defined($self->{'docfiledata'}{$f}{'details'})) {
-        print $fd ' details="', $self->{'docfiledata'}{$f}{'details'}, '"';
+        my $tmp = $self->{'docfiledata'}{$f}{'details'};
+        #$tmp =~ s/\"/\\\"/g;
+        print $fd ' details="', $tmp, '"';
       }
       if (defined($self->{'docfiledata'}{$f}{'language'})) {
-        print $fd ' language="', $self->{'docfiledata'}{$f}{'language'}, '"';
+        my $tmp = $self->{'docfiledata'}{$f}{'language'};
+        #$tmp =~ s/\"/\\\"/g;
+        print $fd ' language="', $tmp, '"';
       }
       print $fd "\n";
     }
@@ -324,6 +343,9 @@ sub writeout {
   if (defined($self->{'srccontainermd5'})) {
     print $fd "srccontainermd5 $self->{'srccontainermd5'}\n";
   }
+  if (defined($self->{'srccontainerchecksum'})) {
+    print $fd "srccontainerchecksum $self->{'srccontainerchecksum'}\n";
+  }
   if (defined($self->{'srcfiles'}) && (@{$self->{'srcfiles'}})) {
     print $fd "srcfiles size=$self->{'srcsize'}\n";
     foreach (sort @{$self->{'srcfiles'}}) {
@@ -398,6 +420,58 @@ sub writeout_simple {
   }
 }
 
+sub as_json {
+  my $self = shift;
+  my %addargs = @_;
+  my %foo = %{$self};
+  # set the additional args
+  for my $k (keys %addargs) {
+    if (defined($addargs{$k})) {
+      $foo{$k} = $addargs{$k};
+    } else {
+      delete($foo{$k});
+    }
+  }
+  # make sure numbers are encoded as numbers
+  for my $k (qw/revision runsize docsize srcsize containersize lrev rrev
+                srccontainersize doccontainersize runcontainersize/) {
+    $foo{$k} += 0 if exists($foo{$k});
+  }
+  for my $k (keys %{$foo{'binsize'}}) {
+    $foo{'binsize'}{$k} += 0;
+  }
+  # encode boolean as boolean flags
+  if (exists($foo{'relocated'})) {
+    if ($foo{'relocated'}) {
+      $foo{'relocated'} = TeXLive::TLUtils::True();
+    } else {
+      $foo{'relocated'} = TeXLive::TLUtils::False();
+    }
+  }
+  # adjust the docfiles entry to the specification in JSON-formats
+  my @docf = $self->docfiles;
+  my $dfd = $self->docfiledata;
+  my @newdocf;
+  for my $f ($self->docfiles) {
+    my %newd;
+    $newd{'file'} = $f;
+    if (defined($dfd->{$f})) {
+      # TODO should we check that there are actually only "details"
+      # and "language" as key?
+      for my $k (keys %{$dfd->{$f}}) {
+        $newd{$k} = $dfd->{$f}->{$k};
+      }
+    }
+    push @newdocf, \%newd;
+  }
+  $foo{'docfiles'} = [ @newdocf ];
+  delete($foo{'docfiledata'});
+  #
+  my $utf8_encoded_json_text = TeXLive::TLUtils::encode_json(\%foo);
+  return $utf8_encoded_json_text;
+}
+
+
 sub cancel_reloc_prefix {
   my $self = shift;
   my @docfiles = $self->docfiles;
@@ -512,7 +586,7 @@ sub make_container {
   # load Cwd only if necessary ...
   require Cwd;
   my $cwd = &Cwd::getcwd;
-  if ("$destdir" !~ m@^(.:)?/@) {
+  if ("$destdir" !~ m@^(.:)?[/\\]@) {
     # we have an relative containerdir, so we have to make it absolute
     $destdir = "$cwd/$destdir";
   }
@@ -580,8 +654,9 @@ sub make_container {
   # overflow standard tar format and result in special things being
   # done.  We don't want the GNU-specific special things.
   #
+  my $is_user_container = ( $containername =~ /\.r[0-9]/ );
   my @attrs
-    = $containername =~ /\.r[0-9]/
+    = $is_user_container
       ? ()
       : ( "--owner", "0",  "--group", "0",  "--exclude", ".svn",
           "--format", "ustar" );
@@ -603,7 +678,7 @@ sub make_container {
   my $tartempfile = "";
   if (win32()) {
     # Since we provide our own (GNU) tar on Windows, we know it has -T.
-    my $tmpdir = TeXLive::TLUtils::get_system_tmpdir();
+    my $tmpdir = TeXLive::TLUtils::tl_tmpdir();
     $tartempfile = "$tmpdir/mc$$";
     open(TMP, ">$tartempfile") || die "open(>$tartempfile) failed: $!";
     print TMP map { "$_\n" } @files_to_backup;
@@ -661,7 +736,13 @@ sub make_container {
     return (0, 0, "");
   }
   my $size = (stat "$destdir/$containername") [7];
-  my $md5 = TeXLive::TLUtils::tlmd5("$destdir/$containername");
+  #
+  # if we are creating a system container, or there is a way to
+  # compute the checksums, do it
+  my $checksum = "";
+  if (!$is_user_container || $::checksum_method) {
+    $checksum = TeXLive::TLCrypto::tlchecksum("$destdir/$containername");
+  }
   
   # cleaning up
   unlink("$tlpobjdir/$self->{'name'}.tlpobj");
@@ -670,8 +751,8 @@ sub make_container {
   rmdir($InfraLocation) if $removetlpkgdir;
   xchdir($cwd);
 
-  debug(" done $containername, size $size, $md5\n");
-  return ($size, $md5, "$destdir/$containername");
+  debug(" done $containername, size $size, $checksum\n");
+  return ($size, $checksum, "$destdir/$containername");
 }
 
 
@@ -707,8 +788,7 @@ sub total_size {
 # Update the current TLPOBJ object with the information from the
 # corresponding entry in C<$tlc->entries>.
 #
-sub update_from_catalogue
-{
+sub update_from_catalogue {
   my ($self, $tlc) = @_;
   my $tlcname = $self->name;
   if (defined($self->catalogue)) {
@@ -731,7 +811,7 @@ sub update_from_catalogue
       $foo =~ s/^.Date: //;
       # trying to extract the interesting part of a subversion date
       # keyword expansion here, e.g.,
-      # $Date: 2013-04-13 07:32:44 +0200 (Sat, 13 Apr 2013) $
+      # $Date: 2018-02-26 19:16:54 +0100 (Mon, 26 Feb 2018) $
       # ->2007-08-15 19:43:35 +0100
       $foo =~ s/ \(.*\)( *\$ *)$//;  # maybe nothing after parens
       $self->cataloguedata->{'date'} = $foo;
@@ -745,6 +825,17 @@ sub update_from_catalogue
     if (defined($entry->ctan) && $entry->ctan ne "") {
       $self->cataloguedata->{'ctan'} = $entry->ctan;
     }
+    # TODO TODO TODO
+    # we should rewrite the also fields to TeX Live package names ...
+    # for now these are CTAN package names!
+    # warning, we expect that cataloguedata entries are strings, 
+    # so stringify these lists
+    if (@{$entry->also}) {
+      $self->cataloguedata->{'also'} = "@{$entry->also}";
+    }
+    if (@{$entry->topics}) {
+      $self->cataloguedata->{'topics'} = "@{$entry->topics}";
+    }
     #if (defined($entry->texlive)) {
     # $self->cataloguedata->{'texlive'} = $entry->texlive;
     #}
@@ -769,24 +860,39 @@ sub update_from_catalogue
     # <documentation details='Package documentation' language='de'
     #   href='ctan:/macros/latex/contrib/juramisc/doc/jmgerdoc.pdf'/>
     # <ctan path='/macros/latex/contrib/juramisc'/>
-    my @tcdocfiles = keys %{$entry->docs};
-    my @tlpdocfiles = $self->docfiles;
-    foreach my $tcdocfile (@tcdocfiles) {
-      # basename also kills the ctan: prefix!
+    my @tcdocfiles = keys %{$entry->docs};  # Catalogue doc files.
+    my %tcdocfilebasenames;                 # basenames of those, as we go.
+    my @tlpdocfiles = $self->docfiles;      # TL doc files.
+    foreach my $tcdocfile (sort @tcdocfiles) {  # sort so shortest first
+      #warn "looking at tcdocfile $tcdocfile\n";
       my $tcdocfilebasename = $tcdocfile;
-      # remove the ctan: prefix
-      $tcdocfilebasename =~ s/^ctan://;
-      # remove the ctan path if present
-      my $tcctanpath = $entry->ctan ? $entry->ctan : "";
-      $tcdocfilebasename =~ s/^$tcctanpath//;
+      $tcdocfilebasename =~ s/^ctan://;  # remove ctan: prefix
+      $tcdocfilebasename =~ s,.*/,,;     # remove all but the base file name
+      #warn "  got basename $tcdocfilebasename\n";
+      #
+      # If we've already seen this basename, skip.  This is for the sake
+      # of README files, which can exist in different directories but
+      # get renamed into different files in TL for various annoying reasons;
+      # e.g., ibygrk, rsfs, songbook.  In these cases, it turns out we
+      # always prefer the first entry (top-level README).
+      next if exists $tcdocfilebasenames{$tcdocfilebasename};
+      $tcdocfilebasenames{$tcdocfilebasename} = 1;
+      #
       foreach my $tlpdocfile (@tlpdocfiles) {
-        if ($tlpdocfile =~ m/$tcdocfilebasename$/) {
-          # update the language/detail tags if present!
+        #warn "considering merge into tlpdocfile $tlpdocfile\n";
+        if ($tlpdocfile =~ m,/$tcdocfilebasename$,) {
+          # update the language/detail tags from Catalogue if present.
           if (defined($entry->docs->{$tcdocfile}{'details'})) {
-            $self->{'docfiledata'}{$tlpdocfile}{'details'} = $entry->docs->{$tcdocfile}{'details'};
+            my $tmp = $entry->docs->{$tcdocfile}{'details'};
+            #warn "merging details for $tcdocfile: $tmp\n";
+            # remove all embedded quotes, they are just a pain
+            $tmp =~ s/"//g;
+            $self->{'docfiledata'}{$tlpdocfile}{'details'} = $tmp;
           }
           if (defined($entry->docs->{$tcdocfile}{'language'})) {
-            $self->{'docfiledata'}{$tlpdocfile}{'language'} = $entry->docs->{$tcdocfile}{'language'};
+            my $tmp = $entry->docs->{$tcdocfile}{'language'};
+            #warn "merging lang for $tcdocfile: $tmp\n";
+            $self->{'docfiledata'}{$tlpdocfile}{'language'} = $tmp;
           }
         }
       }
@@ -1088,6 +1194,18 @@ sub _parse_hyphen_execute {
 
 # member access functions
 #
+sub _set_get_array_value {
+  my $self = shift;
+  my $key = shift;
+  if (@_) { 
+    if (defined($_[0])) {
+      $self->{$key} = [ @_ ];
+    } else {
+      $self->{$key} = [ ];
+    }
+  }
+  return @{ $self->{$key} };
+}
 sub name {
   my $self = shift;
   if (@_) { $self->{'name'} = shift }
@@ -1124,9 +1242,7 @@ sub catalogue {
   return $self->{'catalogue'};
 }
 sub srcfiles {
-  my $self = shift;
-  if (@_) { $self->{'srcfiles'} = [ @_ ] }
-  return @{ $self->{'srcfiles'} };
+  _set_get_array_value(shift, "srcfiles", @_);
 }
 sub containersize {
   my $self = shift;
@@ -1146,17 +1262,47 @@ sub doccontainersize {
 sub containermd5 {
   my $self = shift;
   if (@_) { $self->{'containermd5'} = shift }
-  return ( defined($self->{'containermd5'}) ? $self->{'containermd5'} : "" );
+  if (defined($self->{'containermd5'})) {
+    return ($self->{'containermd5'});
+  } else {
+    tlwarn("TLPOBJ: MD5 sums are no longer supported, please adapt your code!\n");
+    return ("");
+  }
 }
 sub srccontainermd5 {
   my $self = shift;
   if (@_) { $self->{'srccontainermd5'} = shift }
-  return ( defined($self->{'srccontainermd5'}) ? $self->{'srccontainermd5'} : "" );
+  if (defined($self->{'srccontainermd5'})) {
+    return ($self->{'srccontainermd5'});
+  } else {
+    tlwarn("TLPOBJ: MD5 sums are no longer supported, please adapt your code!\n");
+    return ("");
+  }
 }
 sub doccontainermd5 {
   my $self = shift;
   if (@_) { $self->{'doccontainermd5'} = shift }
-  return ( defined($self->{'doccontainermd5'}) ? $self->{'doccontainermd5'} : "" );
+  if (defined($self->{'doccontainermd5'})) {
+    return ($self->{'doccontainermd5'});
+  } else {
+    tlwarn("TLPOBJ: MD5 sums are no longer supported, please adapt your code!\n");
+    return ("");
+  }
+}
+sub containerchecksum {
+  my $self = shift;
+  if (@_) { $self->{'containerchecksum'} = shift }
+  return ( defined($self->{'containerchecksum'}) ? $self->{'containerchecksum'} : "" );
+}
+sub srccontainerchecksum {
+  my $self = shift;
+  if (@_) { $self->{'srccontainerchecksum'} = shift }
+  return ( defined($self->{'srccontainerchecksum'}) ? $self->{'srccontainerchecksum'} : "" );
+}
+sub doccontainerchecksum {
+  my $self = shift;
+  if (@_) { $self->{'doccontainerchecksum'} = shift }
+  return ( defined($self->{'doccontainerchecksum'}) ? $self->{'doccontainerchecksum'} : "" );
 }
 sub srcsize {
   my $self = shift;
@@ -1176,9 +1322,7 @@ sub remove_srcfiles {
   $self->remove_files("src",@files);
 }
 sub docfiles {
-  my $self = shift;
-  if (@_) { $self->{'docfiles'} = [ @_ ] }
-  return @{ $self->{'docfiles'} };
+  _set_get_array_value(shift, "docfiles", @_);
 }
 sub clear_docfiles {
   my $self = shift;
@@ -1234,9 +1378,7 @@ sub remove_binfiles {
   $self->{'binfiles'}{$arch} = [ @finalfiles ];
 }
 sub runfiles {
-  my $self = shift;
-  if (@_) { $self->{'runfiles'} = [ @_ ] }
-  return @{ $self->{'runfiles'} };
+  _set_get_array_value(shift, "runfiles", @_);
 }
 sub clear_runfiles {
   my $self = shift;
@@ -1256,19 +1398,13 @@ sub remove_runfiles {
   $self->remove_files("run",@files);
 }
 sub depends {
-  my $self = shift;
-  if (@_) { $self->{'depends'} = [ @_ ] }
-  return @{ $self->{'depends'} };
+  _set_get_array_value(shift, "depends", @_);
 }
 sub executes {
-  my $self = shift;
-  if (@_) { $self->{'executes'} = [ @_ ] }
-  return @{ $self->{'executes'} };
+  _set_get_array_value(shift, "executes", @_);
 }
 sub postactions {
-  my $self = shift;
-  if (@_) { $self->{'postactions'} = [ @_ ] }
-  return @{ $self->{'postactions'} };
+  _set_get_array_value(shift, "postactions", @_);
 }
 sub containerdir {
   my @self = shift;
@@ -1282,6 +1418,7 @@ sub cataloguedata {
   return $self->{'cataloguedata'};
 }
 
+$: = " \n"; # don't break at -
 format multilineformat =
 longdesc ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<~~
 $_tmp
@@ -1439,13 +1576,13 @@ there are 6 more possible keys:
   $tlpobj->containersize
   $tlpobj->doccontainersize
   $tlpobj->srccontainersize
-  $tlpobj->containermd5
-  $tlpobj->doccontainermd5
-  $tlpobj->srccontainermd5
+  $tlpobj->containerchecksum
+  $tlpobj->doccontainerchecksum
+  $tlpobj->srccontainerchecksum
 
-describing the respective sizes and md5sums in bytes and as hex string, resp.
-The latter two are only present
-if src/doc file container splitting is activated for that install medium.
+describing the respective sizes and checksums in bytes and as hex string, resp.
+The latter two are only present if src/doc file container splitting is
+activated for that install medium.
 
 =head1 OTHER FUNCTIONS
 
@@ -1489,6 +1626,10 @@ or the filehandle if given:
 
 debugging function for comparison with C<tpm>/C<tlps>, will go away.
 
+=item C<as_json>
+
+returns the representation of the C<TLPOBJ> in JSON format.
+
 =item C<common_texmf_tree>
 
 if all files of the package are from the same texmf tree, this tree 
@@ -1524,7 +1665,7 @@ is placed into the root of the installation.
 This is used to distribute packages which can be installed in any arbitrary
 texmf tree (of other distributions, too).
 
-Return values are the size, the md5sum, and the full name of the container.
+Return values are the size, the checksum, and the full name of the container.
 
 =item C<recompute_sizes($tltree)>
 
@@ -1648,8 +1789,10 @@ lines for language.dat.lua that can be generated from the tlpobj.
 
 =head1 SEE ALSO
 
-The modules L<TeXLive::TLConfig>, L<TeXLive::TLUtils>, L<TeXLive::TLPSRC>,
-L<TeXLive::TLPDB>, L<TeXLive::TLTREE>, L<TeXLive::TeXCatalogue>.
+The modules L<TeXLive::TLConfig>, L<TeXLive::TLCrypto>,
+L<TeXLive::TLUtils>, L<TeXLive::TLPSRC>, L<TeXLive::TLPDB>,
+L<TeXLive::TLTREE>, L<TeXLive::TeXCatalogue>, etc., and the
+documentation in the repository: C<Master/tlpkg/doc/>.
 
 =head1 AUTHORS AND COPYRIGHT
 
