@@ -1,13 +1,13 @@
-# $Id: TLPSRC.pm 26458 2012-05-16 23:22:30Z karl $
+# $Id: TLPSRC.pm 44293 2017-05-11 18:08:47Z karl $
 # TeXLive::TLPSRC.pm - module for handling tlpsrc files
-# Copyright 2007, 2008, 2009, 2010, 2011, 2012 Norbert Preining
+# Copyright 2007-2017 Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
 package TeXLive::TLPSRC;
 
 use FileHandle;
-use TeXLive::TLConfig qw($CategoriesRegexp);
+use TeXLive::TLConfig qw($CategoriesRegexp $DefaultCategory);
 use TeXLive::TLUtils;
 use TeXLive::TLPOBJ;
 use TeXLive::TLTREE;
@@ -15,7 +15,7 @@ use TeXLive::TLTREE;
 my $_tmp;
 my %autopatterns;  # computed once internally
 
-my $svnrev = '$Revision: 26458 $';
+my $svnrev = '$Revision: 44293 $';
 my $_modulerevision;
 if ($svnrev =~ m/: ([0-9]+) /) {
   $_modulerevision = $1;
@@ -83,6 +83,7 @@ sub from_file {
   my $foundnametag = 0;
   my $finished = 0;
   my $savedline = "";
+  my %tlpvars;
 
   my $lineno = 0;
   for my $line (@lines) {
@@ -105,7 +106,7 @@ sub from_file {
     # (blank lines are significant in tlpobj, but not tlpsrc)
 
     if ($line =~ /^ /) {
-      die "$srcfile: non-continuation indentation not allowed: `$line'";
+      die "$srcfile:$lineno: non-continuation indentation not allowed: `$line'";
     }
     # remove terminal white space
     $line =~ s/\s+$//;
@@ -119,6 +120,12 @@ sub from_file {
       $foundnametag && die "$srcfile: second name directive not allowed: $name";
       $foundnametag = 1;
     } else {
+      # expand tlpvars while reading in
+      # that means we have to respect *order* and define variables
+      # first in the tlpsrc file
+      for my $k (keys %tlpvars) {
+        $line =~ s/\$\{\Q$k\E\}/$tlpvars{$k}/g;
+      }
       # we default to the file name as package name
       # $started || die "$srcfile: first directive must be `name', not $line";
       if ($line =~ /^shortdesc\s*(.*)$/) {
@@ -160,6 +167,9 @@ sub from_file {
       } elsif ($line =~ /^postaction\s+(.*)$/) {
         push @postactions, $1 if ($1 ne "");
         next;
+      } elsif ($line =~ /^tlpsetvar\s+([-_a-zA-Z0-9]+)\s+(.*)$/) {
+        $tlpvars{$1} = $2;
+        next;
       } else {
         tlwarn("$srcfile:$lineno: unknown tlpsrc directive, fix fix: $line\n");
       }
@@ -187,7 +197,7 @@ sub from_file {
 sub writeout {
   my $self = shift;
   my $fd = (@_ ? $_[0] : STDOUT);
-  format_name $fd "multilineformat";
+  format_name $fd "multilineformat";  # format defined in TLPOBJ, and $:
   $fd->format_lines_per_page (99999); # no pages in this format
   print $fd "name ", $self->name, "\n";
   print $fd "category ", $self->category, "\n";
@@ -252,6 +262,24 @@ sub make_tlpobj {
   $tlp->postactions(@{$self->{'postactions'}}) if (defined($self->{'postactions'}));
   $tlp->depends(@{$self->{'depends'}}) if (defined($self->{'depends'}));
   $tlp->revision(0);
+
+  # convert each fmttrigger to a depend line, if not already present.
+  if (defined($tlp->executes)) { # else no fmttriggers
+    my @deps = (defined($tlp->depends) ? $tlp->depends : ());
+    for my $e ($tlp->executes) {
+      # we only check for AddFormat lines
+      if ($e =~ m/^\s*AddFormat\s+(.*)\s*$/) {
+        my %fmtline = TeXLive::TLUtils::parse_AddFormat_line($1);
+        if (defined($fmtline{"error"})) {
+          tlwarn ("error in parsing $e for return hash: $fmtline{error}\n");
+        } else {
+          TeXLive::TLUtils::push_uniq (\@deps, @{$fmtline{'fmttriggers'}});
+          $tlp->depends(@deps);
+        }
+      }
+    }
+  }
+
   my $filemax;
   my $usedefault;
   my @allpospats;
@@ -420,7 +448,8 @@ sub make_tlpobj {
       if (!@archfiles) {
         if (($arch ne "win32") || defined($::tlpsrc_pattern_warn_win)) {
           tlwarn("$self->{name} ($arch): no hit on negative binpattern $finalp\n")
-            unless defined($::tlpsrc_pattern_no_warn_negative);
+            unless $::tlpsrc_pattern_no_warn_negative;
+            # see comments in libexec/place script.
         }
       }
       $tlp->remove_binfiles($arch,@archfiles);
@@ -443,7 +472,8 @@ sub _do_normal_pattern {
   my @matchfiles = $tltree->get_matching_files($type, $p, $self->{'name'});
   if (!$is_default_pattern && !@matchfiles
       && ($p !~ m,^f ignore,) && ($p !~ m,^d tlpkg/backups,)) {
-    tlwarn("$self->{name}: no hit for pattern $p\n");
+    tlwarn("$self->{name}: no hit for pattern $p\n")
+      unless $negative && $::tlpsrc_pattern_no_warn_negative;
   }
   if (defined($negative) && $negative == 1) {
     $tlp->remove_files($type,@matchfiles);
@@ -598,11 +628,6 @@ sub postactions {
   return @{ $self->{'postactions'} };
 }
 
-format multilineformat =
-longdesc ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<~~
-$_tmp
-.
-
 1;
 __END__
 
@@ -850,6 +875,13 @@ instead of the one given via C<file>.
 
 =back
 
+=item C<tlpsetvar> I<var> I<val>
+
+sets variable I<var> to I<val>, for use within the current C<.tlpsrc>
+only.  Order matters.  The variable can be expanded with
+C<${>I<var>C<}> (after it is defined).  Characters allowed in the I<var>
+name are C<-_a-zA-Z0-9>.
+
 =item C<(src|run|doc|bin)pattern> I<pattern>
 
 adds a pattern (next section) to the respective list of patterns.
